#version 150
precision highp float;

//Ethan Alexander Shulman "Mandel Sphere Fractal" November 4, 2020

in vec3 in_Position;
out float time;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
#define screen vec2(screenX,screenY)

#define ANIMATION 1
#define FSAMPLES 1000

#define STEPS 100
#define EPS 4e-2
#define PI 3.14159265

mat2 r2(float a) {
	float s = sin(a), c = cos(a);
	return mat2(c,-s,s,c);
}

mat2x3 tangentBinormal(vec3 norm) {
	mat2x3 t;
	if (abs(norm.x) > abs(norm.y)) t[0] = normalize(vec3(-norm.z,0,norm.x));
	else t[0] = normalize(vec3(0,norm.z,-norm.y)); 
	t[1] = cross(norm,t[0]);
	return t;
}


vec3 hash(uvec3 x) {
    x = ((x>>8U)^x.yzx)*1103515245U;
    x = ((x>>8U)^x.yzx)*1103515245U;
    x = ((x>>8U)^x.yzx)*1103515245U;
    
    return vec3(x)*(1.0/float(0xffffffffU));
}

vec3 hsvToRGB(vec3 hsv) {
	return mix(clamp(abs(mod(hsv.x*6.0+vec3(0.,4.,2.),6.)-3.0)-1.0,0.,1.),vec3(1),hsv.y)*hsv.z;
}

#define START_OPACITY(posSeed) vec3 opacityRand = hash(uvec3(abs(posSeed)*1000.)+uvec3(frame,frame*3,frame*7))
#define OPACITY(dst,skip,alpha) (dst+skip*opacityRand.y*step(alpha,opacityRand.x))

float geometry(vec3 p) {
	float d = 1e8, sz = 5.;
	for (float i = 1.; i < 8.; i++) {
		d = min(d,length(p)-sz);
		p = abs(abs(p)-sz);
		sz *= .45;
	}
	return d;
}
//mat[0] = color, mat[1].x = roughness, mat[1].y = emission
mat2x3 material(vec3 p) {
	
	return mat2x3(vec3(.5), vec3(.1,0.,0));
}
vec3 normal(vec3 p) {
	#define NEPS vec3(EPS,0,0)
	vec3 n = normalize(vec3(geometry(p+NEPS),geometry(p+NEPS.yxz),geometry(p+NEPS.yzx))-geometry(p));
	return n!=n?normalize(fract(abs(p.zxy*1e6))-.5):n;
}

vec3 background(vec3 rp, vec3 rd) {
	return vec3(.2,.4,.7)+vec3(1)*max(0.,1.-length(rd-vec3(-1,1,-1)))*200.;
}

float twave(float x) {
	return 1.-abs(fract(x)-.5)*2.;
}

uniform sampler2D Buffer0;

void vert() {
	time = float(frame)*.25/float(framerate*FSAMPLES);
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*screen);
	gl_Position = vec4(in_Position,1);
}
void frag() {
	vec3 rp = vec3(0,-7.,0.),//camera position
		rd = normalize(vec3(uv.xy+(hash(uvec3(uv.zw,frame)).xy*4.-2.)/screenY,-.5)),
		c = vec3(1);
	float l = 0.;
	
	//camera rotation
	mat2 pitch = r2(time*PI);
	rp.yz *= pitch;
	rd.yz *= pitch;

	int i;
	for (i = 0; i < STEPS; i++) {
		float dst = geometry(rp);
		if (dst <= EPS) {
			mat2x3 mat = material(rp);
			l += mat[1].y;
			c *= mat[0];
			
			rp -= rd*(EPS*2.-dst);
			vec3 nrm = normal(rp);
			mat2x3 tangBinorm = tangentBinormal(nrm);
			
			vec2 rnd = hash(uvec3(abs(rp)*1000.)+uvec3(frame,frame*3,frame*7)).xy;
			float ct = pow(rnd.x,mat[1].x),
				st = sqrt(1.-ct*ct),
				pt = rnd.y*PI*2.;
			vec3 surf = normalize(nrm*ct+tangBinorm[0]*(st*cos(pt))+tangBinorm[1]*(st*sin(pt)));
			rd = reflect(rd,surf);
			dst = 0.;
		}
		rp += rd*dst;
		
		if (dot(rp,rp) > 20.*20.) i = STEPS;
	}
	
	//background + final lighting
	c *= l+background(rp,rd)*(i>=STEPS?1.:clamp(geometry(rp)-5.,0.,1.));
	
	#if ANIMATION
	gl_FragColor = (frame%FSAMPLES==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+vec4(c,1)*1e-2;
	#else
	gl_FragColor = (frame==0?vec4(0):texelFetch(Buffer0,ivec2(uv.zw),0))+vec4(c,1)*1e-2;
	#endif
}
